/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package components

import (
	"fmt"
	"strconv"
	"strings"

	corev1 "k8s.io/api/core/v1"
	networkingv1beta1 "k8s.io/api/networking/v1beta1"
	"k8s.io/apimachinery/pkg/util/intstr"

	"k8s.libre.sh/interfaces"
	meta "k8s.libre.sh/meta"
	ingress "k8s.libre.sh/objects/ingress"
	service "k8s.libre.sh/objects/service"
)

// +kubebuilder:object:generate=true
type Network struct {
	IngressMeta *meta.ObjectMeta `json:"ingressMeta,omitempty"`
	ServiceMeta *meta.ObjectMeta `json:"serviceMeta,omitempty"`
	*Backend    `json:",inline"`
}

func (s *Network) TemplateValues() map[string]interface{} {
	values := make(map[string]interface{})
	values["backend"] = s.Backend.TemplateValues()
	values["serviceMeta"] = s.ServiceMeta.TemplateValues()
	values["ingressMeta"] = s.IngressMeta.TemplateValues()

	return values
}

// +kubebuilder:object:generate=true
type Service struct {
	*meta.ObjectMeta `json:"meta,omitempty"`
	*Backend         `json:"backend,omitempty"`
}

func (s *Service) Mutate(obj interfaces.Object) error {
	service.MutateService(s, obj.(*corev1.Service), s.ObjectMeta)
	meta.MutateMeta(s, obj)
	return nil
}

func (s *Service) GetObject() interfaces.Object {
	return &corev1.Service{}
}

func (s *Service) TemplateValues() map[string]interface{} {
	values := make(map[string]interface{})
	values["meta"] = s.ObjectMeta.TemplateValues()
	values["paths"] = s.Paths
	values["port"] = map[string]string{
		"name":     s.Port.Name,
		"port":     strconv.FormatInt(int64(s.Port.Port), 10),
		"protocol": string(s.Port.Protocol),
	}
	values["serviceType"] = s.ServiceType
	return values
}

// +kubebuilder:object:generate=true
type Ingress struct {
	*meta.ObjectMeta `json:"commonMeta,omitempty"`
	Backends         []Backend `json:"backends,omitempty"`
}

// +kubebuilder:object:generate=true
type Backend struct {
	ServiceName  string             `json:"name,omitempty"`
	Hostname     string             `json:"hostname,omitempty"`
	Paths        []string           `json:"routes,omitempty"`
	TLS          bool               `json:"tls,omitempty"`
	TLSSecretRef string             `json:"tlsSecretRef,omitempty"`
	Port         service.Port       `json:"port,omitempty"`
	ServiceType  corev1.ServiceType `json:"type,omitempty"`
}

func (s *Backend) TemplateValues() map[string]interface{} {
	values := make(map[string]interface{})
	values["serviceName"] = s.ServiceName
	values["Hostname"] = s.Hostname
	values["Paths"] = s.Paths
	values["TLS"] = s.TLS
	values["TLSSecretRef"] = s.TLSSecretRef
	values["port"] = map[string]string{
		"name":     s.Port.Name,
		"port":     strconv.FormatInt(int64(s.Port.Port), 10),
		"protocol": string(s.Port.Protocol),
	}
	values["ServiceType"] = s.ServiceType

	return values
}

func (i *Ingress) Mutate(obj interfaces.Object) error {
	ingress.MutateIngress(i, obj.(*networkingv1beta1.Ingress))
	meta.MutateMeta(i, obj)
	return nil
}

func (i *Ingress) GetObject() interfaces.Object {
	return &networkingv1beta1.Ingress{}
}

func (b *Backend) GetServiceType() corev1.ServiceType { return b.ServiceType }

func (b *Backend) GetServicePorts() []corev1.ServicePort {

	ports := []corev1.ServicePort{}
	ports = append(ports, b.GetServicePort())

	return ports
}

func (b *Backend) GetServicePort() corev1.ServicePort {

	port := corev1.ServicePort{
		TargetPort: intstr.FromString(b.Port.Name),
		Port:       b.Port.Port,
		Name:       b.Port.Name,
	}

	if len(b.Port.Protocol) > 0 {
		port.Protocol = b.Port.Protocol
	}

	return port
}

func (b *Backend) GetIngressBackendPaths() []networkingv1beta1.HTTPIngressPath {

	bkpaths := []networkingv1beta1.HTTPIngressPath{}

	for _, p := range b.Paths {
		path := networkingv1beta1.HTTPIngressPath{
			Path: p,
			Backend: networkingv1beta1.IngressBackend{
				// TODO TO FIX
				ServiceName: b.ServiceName,
				ServicePort: intstr.FromString(b.Port.Name),
			},
		}

		bkpaths = append(bkpaths, path)
	}
	return bkpaths
}

func (b *Backend) GetIngressRule() networkingv1beta1.IngressRule {

	rule := networkingv1beta1.IngressRule{
		Host: b.Hostname,
		IngressRuleValue: networkingv1beta1.IngressRuleValue{
			HTTP: &networkingv1beta1.HTTPIngressRuleValue{
				Paths: b.GetIngressBackendPaths(),
			},
		},
	}

	return rule
}

func (i *Ingress) GetIngressRules() []networkingv1beta1.IngressRule {

	rules := []networkingv1beta1.IngressRule{}

	for _, ing := range i.Backends {
		rules = append(rules, ing.GetIngressRule())
	}

	return rules
}

func (b *Backend) GetIngressTLS() networkingv1beta1.IngressTLS {
	tls := networkingv1beta1.IngressTLS{}
	if b.TLS {
		if len(b.TLSSecretRef) == 0 {
			b.TLSSecretRef = fmt.Sprintf("%v-%v", strings.Replace(b.Hostname, ".", "-", -1), "tls")
		}

		tls = networkingv1beta1.IngressTLS{
			SecretName: b.TLSSecretRef,
		}

		tls.Hosts = append(tls.Hosts, b.Hostname)
	}
	return tls
}

func (i *Ingress) GetIngressTLS() []networkingv1beta1.IngressTLS {

	tlsList := []networkingv1beta1.IngressTLS{}

	for _, ing := range i.Backends {
		tlsList = append(tlsList, ing.GetIngressTLS())
	}

	return tlsList
}
